;*---------------------------------------------------------------------*/
;*    Copyright (c) 1994 by Manuel Serrano. All rights reserved.       */
;*                                                                     */
;*                                     ,--^,                           */
;*                               _ ___/ /|/                            */
;*                           ,;'( )__, ) '                             */
;*                          ;;  //   L__.                              */
;*                          '   \    /  '                              */
;*                               ^   ^                                 */
;*                                                                     */
;*                                                                     */
;*    This program is distributed in the hope that it will be useful.  */
;*    Use and copying of this software and preparation of derivative   */
;*    works based upon this software are permitted, so long as the     */
;*    following conditions are met:                                    */
;*           o credit to the authors is acknowledged following         */
;*             current academic behaviour                              */
;*           o no fees or compensation are charged for use, copies,    */
;*             or access to this software                              */
;*           o this copyright notice is included intact.               */
;*      This software is made available AS IS, and no warranty is made */
;*      about the software or its performance.                         */
;*                                                                     */
;*      Bug descriptions, use reports, comments or suggestions are     */
;*      welcome Send them to                                           */
;*        <Manuel.Serrano@inria.fr>                                    */
;*        Manuel Serrano                                               */
;*        INRIA -- Rocquencourt                                        */
;*        Domaine de Voluceau, BP 105                                  */
;*        78153 Le Chesnay Cedex                                       */
;*        France                                                       */
;*---------------------------------------------------------------------*/


;*=====================================================================*/
;*    .../pmakeblock.scm ...                                           */
;*    -------------------------------------------------------------    */
;*    Author      :  Manuel Serrano                                    */
;*    Creation    :  Thu Nov 10 17:15:08 1994                          */
;*    Last change :  Thu Dec  8 09:20:58 1994 (serrano)                */
;*    -------------------------------------------------------------    */
;*    La compilation de la forme `Pmakeblock'                          */
;*=====================================================================*/

;*---------------------------------------------------------------------*/
;*    Le module                                                        */
;*---------------------------------------------------------------------*/
(module pmakeblock
   (include "Camloo/trace.sch")
   (import  misc
	    llambda
	    lconst)
   (export (Pmakeblock <expr> <constr> <values> <cp> <bool> <bool>)))

;*---------------------------------------------------------------------*/
;*    Pmakeblock ...                                                   */
;*---------------------------------------------------------------------*/
(define (Pmakeblock exp constr values cp sg? raise?)
   (trace 12 "Pmakeblock: " (shape constr) (shape values))
   (match-case constr
      ((ConstrRegular (qualifiedident ?module ?name) ?- ?-)
       (cond
	  ((and (string=? module "builtin")
		(string=? name "::"))
	   (if (and sg? (constant-expressions? values))
	       (Lconst exp)
	       `(cons ,@(map (lambda (v) (llambda v #f cp #f #f #f)) values))))
	  ((and (string=? module "ref")
		(string=? name "ref"))
	   `(make-cell ,(llambda (car values) #f cp #f #f #f)))
	  (else
	   (if (and sg? (constant-expressions? values))
	       ;; c'est une constante et on la lie a une variable
	       ;; globale, on ne produit pas de code mais une constante.
	       (Lconst exp)
	       ;; ce n'est pas une constante, on produit du code
	       `(,(if (ConstrExtensible? constr)
		      (if raise?
			  'caml-make-static-extensible
			  'caml-make-extensible)
		      'caml-make-regular)
		 ,(constr-tag constr)
		 ,(length values)
		 ,@(map (lambda (v) (llambda v #f cp #f #f #f))
			values))))))
      (else
       (if (and sg? (constant-expressions? values))
	   ;; c'est une constante et on la lie a une variable
	   ;; globale, on ne produit pas de code mais une constante.
	   (Lconst exp)
	   ;; ce n'est pas une constante, on produit du code
	   `(,(if (ConstrExtensible? constr)
		  (if raise?
		      'caml-make-static-extensible
		      'caml-make-extensible)
		  'caml-make-regular)
	     ,(constr-tag constr)
	     ,(length values)
	     ,@(map (lambda (v) (llambda v #f cp #f #f #f))
		    values))))))

;*---------------------------------------------------------------------*/
;*    constant-expressions? ...                                        */
;*---------------------------------------------------------------------*/
(define (constant-expressions? values)
   (trace 12 "constant-expressions? : " (shape values))
   (define (constant-expr? expr)
      (trace 12 "constant-expr? : " (shape expr))
      (match-case expr
	 ((lconst . ?-)
	  #t)
	 ((lprim (pmakeblock ?-) ?values)
	  (constant-expressions? values))
	 (else
	  #f)))
   (let loop ((values values))
      (cond
	 ((null? values)
	  (trace 12 "...   OUI")
	  #t)
	 ((constant-expr? (car values))
	  (loop (cdr values)))
	 (else
	  (trace 12 "...   NON")
	  #f))))

